/*
 * Title:        bits/data_t.hpp
 * Description:  C++ header-only wrapper for C-ABI data_t type
 * $Copyright (c) 2015-2017, 2019, 2022-2023 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 */

/* This file:
 * Contains the C++ implementation of data_t
 *
 */

#ifndef _DATAT_BITS_INCLUDE_TAG
#error Must only be included from sag_connectivity_cpp.hpp, use that instead
#endif

/**
 * A variant type which can be one of the following:
 *
 * - empty
 * - boolean
 * - int64_t
 * - double
 * - const char* (cannot be null)
 * - list_t: list of data_t
 * - map_t: map of data_t:data_t
 * - com::softwareag::connectivity::decimal_t
 * - buffer_t (a byte array)
 * - custom_t (a void* + deleter and copy function pointers)
 *
 * empty is equivalent to a null value in Java.
 *
 * This is moveable, but not implicitly copyable. If you want to copy it use
 * the .copy() method. If you want to move it use std::move().
 *
 * To extract data from this variant you should either use the get<T>() free
 * function if you know which type it is, or use the apply_visitor free
 * function if you do not.
 */
class data_t: protected sag_underlying_data_t
{
	friend class map_t;
	friend class list_t;
	friend class Message;
public:

	/** Construct an empty data_t */
	data_t();
	/** Construct a data_t, move the contents from another data_t.
		Other will be set to the empty data_t */
	data_t(data_t &&other);

	/** Swap the contents of this data_t with another */
	void swap(data_t &&other);

	/** Move-assignment from another data_t.
		Other will be set to the empty data_t and the current contents freed. */
	data_t &operator=(data_t &&other)
	{
		data_t tmp(std::move(other));
		swap(std::move(tmp));
		return *this;
	}

	/** For non-primitives free the underlying memory */
	~data_t();

	/** Returns true if this is the empty data_t */
	bool empty() const { return tag == SAG_DATA_EMPTY; }

	/** Return a deep copy of this data_t */
	data_t copy() const;

	/** Returns true if two data_ts are deep-equals */
	bool operator==(const data_t &other) const;
	/** Returns true if two data_ts are non-equal (deep equality) */
	bool operator!=(const data_t &other) const { return !operator==(other); }

	/**
	 * Apply a visitor (using the boost::static_visitor pattern) to this
	 * data_t.
	 *
	 * This is the main mechanism for accessing the data if you don't know
	 * which type. Your visitor should implement operator() for each of the
	 * possible contained types and the appropriate method will be called,
	 * depending on which type is in the variant. If the variant is empty then
	 * the 0-args operator() will be used.
	 *
	 * The return value of the operators will be returned from this method.
	 * Visitors must define a result_type member (or make use of 
	 * boost::static_visitor, which provides this). Alternatively 
	 * connectivity::visitor may be used when you only need to implement
	 * specific type visitation which also provides result_type.
	 */	 
	template<typename V>
	typename V::result_type apply_visitor(const V &v) const;
	/**
	 * Apply a visitor (using the boost::static_visitor pattern) to this
	 * data_t.
	 *
	 * This is the main mechanism for accessing the data if you don't know
	 * which type. Your visitor should implement operator() for each of the
	 * possible contained types and the appropriate method will be called,
	 * depending on which type is in the variant. If the variant is empty then
	 * the 0-args operator() will be used.
	 *
	 * The return value of the operators will be returned from this method.
	 * Visitors must define a result_type member (or make use of 
	 * boost::static_visitor, which provides this). Alternatively 
	 * connectivity::visitor may be used when you only need to implement
	 * specific type visitation which also provides result_type.
	 */	 
	template<typename V>
	typename V::result_type apply_visitor(const V &v);

	/** Return the name of the contained type as a string (mainly useful for
		debugging)
		
	*/
	const char *type_name() const
	{
		switch (tag) {
			case SAG_DATA_EMPTY: return "empty";
			case SAG_DATA_BOOLEAN: return "boolean";
			case SAG_DATA_DOUBLE: return "double";
			case SAG_DATA_INTEGER: return "integer";
			case SAG_DATA_DECIMAL: return "decimal";
			case SAG_DATA_STRING: 
			case SAG_DATA_CONST_STRING: return "string";
			case SAG_DATA_LIST: return "list";
			case SAG_DATA_MAP: return "map";
			case SAG_DATA_BUFFER: return "buffer";
			case SAG_DATA_CUSTOM: return "custom";
			default:
				assert(false); // shouldn't happen
				return "Unknown";
		}
	}

	/** Return the discriminator for this union (as an enum of the possible
		values) */
	sag_data_tag type_tag() const { return tag==SAG_DATA_CONST_STRING?SAG_DATA_STRING:tag; }

private:
	/** This is not implicitly copiable. Use the copy() method. */
	data_t(const data_t &other) { abort(); }
	/** This is not implicitly copiable. Use the copy() method. */
	data_t &operator=(const data_t &other) { abort(); }
public:

	/**
	 * \private
	 */
	struct do_not_copy_t {};
	
	/** Create a data_t from an int64_t */
	data_t(int64_t d)
	{
		tag = SAG_DATA_INTEGER;
		integer = d;
	}
	/** Assign an int64_t to this, freeing the existing contents of this */
	data_t &operator=(int64_t d)
	{
		data_t tmp(d);
		swap(std::move(tmp));
		return *this;
	}
	/** Create a data_t from a bool */
	explicit data_t(bool b)
	{
		tag = SAG_DATA_BOOLEAN;
		boolean = b;
	}
	/** Assign a bool to this, freeing the existing contents of this */
	data_t &operator=(bool b)
	{
		data_t tmp(b);
		swap(std::move(tmp));
		return *this;
	}
	/** Create a data_t from a double */
	data_t(double d)
	{
		tag = SAG_DATA_DOUBLE;
		fp = d;
	}
	/** Assign a double to this, freeing the existing contents of this */
	data_t &operator=(double d)
	{
		data_t tmp(d);
		swap(std::move(tmp));
		return *this;
	}
	/** Create a data_t from a custom value
		This is a move constructor and will leave the given custom_t empty.
		If you want to copy it into this data_t use custom.copy() Otherwise use
		std::move(). */
	template<typename T>
	data_t(custom_t<T> &&d);

	/** Assign this to a custom value, freeing the existing contents
		This is a move constructor and will leave the given custom_t empty.
		If you want to copy it into this data_t use custom.copy() Otherwise use
		std::move(). */
	template<typename T>
	data_t &operator=(custom_t<T> &&d)
	{
		data_t tmp(std::move(d));
		swap(std::move(tmp));
		return *this;
	}
	/** Create a data_t from a decimal */
	data_t(decimal_t d)
	{
		tag = SAG_DATA_DECIMAL;
		decimal = d;
	}
	/** Assign this to a decimal, freeing the existing contents */
	data_t &operator=(decimal_t d)
	{
		data_t tmp(d);
		swap(std::move(tmp));
		return *this;
	}
	/** Create a data_t from a string (copies the string).
	 * Only use the first n bytes in the string.
		Must be a non-null utf8-encoded string. */
	data_t(const char *s, size_t n);
	/** Create a data_t from a string (copies the string).
		Must be a non-null utf8-encoded string. */
	data_t(const char *s);
	/** Create a data_t from a string without copying it.
	 * Must be a non-null utf8-encoded string. 
	 * \private
	 */
	data_t(const char *s, do_not_copy_t);
	/** Create a data_t from a string (copies the string).
		Must be utf8-encoded string. */
	data_t(const std::string &s);
	/** Assign this to a string, freeing the existing contents (copies the
		string).
		Must be a non-null utf8-encoded string. */
	data_t &operator=(const char *s)
	{
		data_t tmp(s);
		swap(std::move(tmp));
		return *this;
	}

	/** Assign this to a string, freeing the existing contents (copies the
		string).
		Must be utf8-encoded string. */
	data_t &operator=(const std::string &s)
	{
		data_t tmp(s);
		swap(std::move(tmp));
		return *this;
	}
	/** Create a data_t from a list_t
		This is a move constructor and will leave the given list empty.
		If you want to copy it into this data_t use list.copy() Otherwise use
		std::move(). */
	data_t(list_t &&d);
	/** Assign this to a list_t, freeing the existing contents.
		This is a move constructor and will leave the given list empty.
		If you want to copy it into this data_t use list.copy() Otherwise use
		std::move(). */
	data_t &operator=(list_t &&d)
	{
		data_t tmp(std::move(d));
		swap(std::move(tmp));
		return *this;
	}
	/** Create a data_t from a map_t
		This is a move constructor and will leave the given map empty.
		If you want to copy it into this data_t use map.copy() Otherwise use
		std::move(). */
	data_t(map_t &&d);
	/** Assign this to a map_t, freeing the existing contents.
		This is a move constructor and will leave the given map empty.
		If you want to copy it into this data_t use map.copy() Otherwise use
		std::move(). */
	data_t &operator=(map_t &&d)
	{
		data_t tmp(std::move(d));
		swap(std::move(tmp));
		return *this;
	}
	/** Create a data_t from a buffer_t
		This is a move constructor and will leave the given buffer empty.
		If you want to copy it into this data_t use buffer.copy() Otherwise use
		std::move(). */
	data_t(buffer_t &&d);
	/** Assign this to a buffer_t, freeing the existing contents.
		This is a move constructor and will leave the given buffer empty.
		If you want to copy it into this data_t use buffer.copy() Otherwise use
		std::move(). */
	data_t &operator=(buffer_t &&d)
	{
		data_t tmp(std::move(d));
		swap(std::move(tmp));
		return *this;
	}
};
