/*
 * Title:        bits/functions.hpp
 * Description:  C++ header-only wrapper for C-ABI data_t type
 * $Copyright (c) 2015-2017, 2022-2023 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 */

/* This file:
 * Contains free functions relating to data_t
 *
 */

#ifndef _DATAT_BITS_INCLUDE_TAG
#error Must only be included from sag_connectivity_cpp.hpp, use that instead
#endif

#include <sag_internal/convert_to.hpp>

/**
 * Apply a visitor (using the boost::static_visitor pattern) to the
 * given data_t.
 *
 * This is the main mechanism for accessing the data if you don't know
 * which type. Your visitor should implement operator() for each of the
 * possible contained types and the appropriate method will be called,
 * depending on which type is in the variant. If the variant is empty then
 * the 0-args operator() will be used.
 *
 * The return value of the operators will be returned from this method.
 * Visitors must define a result_type member (or make use of 
 * boost::static_visitor, which provides this). Alternatively 
 * connectivity::visitor may be used when you only need to implement
 * specific type visitation which also provides result_type.
 */	 
template<typename V>
inline typename V::result_type apply_visitor(const V&v, data_t& t) { return t.apply_visitor(v); }
/**
 * Apply a visitor (using the boost::static_visitor pattern) to the
 * given data_t.
 *
 * This is the main mechanism for accessing the data if you don't know
 * which type. Your visitor should implement operator() for each of the
 * possible contained types and the appropriate method will be called,
 * depending on which type is in the variant. If the variant is empty then
 * the 0-args operator() will be used.
 *
 * The return value of the operators will be returned from this method.
 * Visitors must define a result_type member (or make use of 
 * boost::static_visitor, which provides this). Alternatively 
 * connectivity::visitor may be used when you only need to implement
 * specific type visitation which also provides result_type.
 */	 
template<typename V>
inline typename V::result_type apply_visitor(const V&v, const data_t& t) { return t.apply_visitor(v); }

/**
 * Get the contents of a data_t, given the type.
 *
 * This is the primary way of accessing the contents of a data_t if you
 * know which type it will be.
 *
 * get is templated over one of the types the data_t can contain. If the
 * data_t contains a value of that type then it will be returned. If not, get
 * will throw std::runtime_error.
 */
template<typename T>
inline typename get_details::GetVisitor<const T>::result_type get(const data_t &t) { return apply_visitor(get_details::GetVisitor<const T>(), t); }
/**
 * Specialization to be a pass-through, for use in a templated context
 */
template<>
inline const data_t &get<const data_t&>(const data_t &t) { return t; }
/**
 * Get the contents of a data_t, given the type.
 *
 * This is the primary way of accessing the contents of a data_t if you
 * know which type it will be.
 *
 * get is templated over one of the types the data_t can contain. If the
 * data_t contains a value of that type then it will be returned. If not, get
 * will throw std::runtime_error.
 */
template<typename T>
inline typename get_details::GetVisitor<T>::result_type get(data_t &t) { return apply_visitor(get_details::GetVisitor<T>(), t); }
/**
 * Specialization to be a pass-through, for use in a templated context
 */
template<>
inline data_t &get<data_t&>(data_t &t) { return t; }
/**
 * Get the contents of the data_t as a typed custom_t.
 *
 * This is a shorthand for using get:
 *
 * get_custom<T>(data) == get<custom_t<T> >(data)
 */
template<typename T>
inline typename get_details::GetVisitor<custom_t<T> >::result_type get_custom(data_t &t) { return apply_visitor(get_details::GetVisitor<custom_t<T> >(), t); }
/**
 * Get the contents of the data_t as a typed custom_t.
 *
 * This is a shorthand for using get:
 *
 * get_custom<T>(data) == get<custom_t<T> >(data)
 */
template<typename T>
inline typename get_details::GetVisitor<const custom_t<T> >::result_type get_custom(const data_t &t) { return apply_visitor(get_details::GetVisitor<const custom_t<T> >(), t); }

/**
 * Get a T from a data_t, parsing it from a string if necessary/possible.
 * If it cannot be converted, or is not either T or a string, then will throw
 * std::runtime_error.
 *
 * Only the following types can be used with convert_to:
 *  - int64_t
 *  - double
 *  - bool
 *  - std::string
 */
template<typename T>
inline T convert_to(const data_t &d) 
{
	// if you get an error like:
	// no match for call to '(const convert_to_details::_get_or_convert<TYPE>) ()'
	// then you're probably trying to use convert_to with an incorrect type. 
	// You can only use it for double, bool or int64_t.
	// For other types use get<TYPE>, but it won't parse it to/from a
	// string.
	return apply_visitor(convert_to_details::_get_or_convert<T>(d.type_name()), d);
}

/**
 * @internal
 * Automatic conversion from C++ connectivity types to equivalent underlying C types, necessary for the public to_string signature.
 */
template<typename T>
struct get_underlying : std::false_type {};

/**
 * Get a string representation of t. Supports data_t, map_t and list_t. Intended only for logging, not performant.
 */
template<typename T>
inline typename std::enable_if<get_underlying<T>::value, std::string>::type to_string(const T &t);

/**
 * Get the length of the object contained in the data_t.
 * @returns 0 if empty, the length of the string, size of the map or list, otherwise 1.
 */
inline size_t length(const data_t &d);
