/*
 * Title:        epl_plugin.hpp
 * Description:  C++ API for writing EPL plug-ins
 * $Copyright (c) 2015-2017, 2020, 2022 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 */

/** @file
 *
 * Contains the headers needed to implement EPL plug-ins in C++. 
 *
 * This re-uses the types defined in sag_connectivity_cpp.hpp and the logger defined 
 * in sag_plugin_logging.hpp. For convenience all of the relevant types and functions
 * have been brought into the com::apama::epl namespace as well.
 */

#ifndef _APAMA_EPL_PLUGINS_HPP_
#define _APAMA_EPL_PLUGINS_HPP_

#include <sag_connectivity_cpp.hpp>
#include <memory>

#include <sag_plugin_logging.hpp>
#include <vector>

namespace com {
namespace apama {
namespace correlator {
class CorrelatorInterface;
} // com.apama.correlator

/**
 * Contains classes needed for implementing EPL plug-ins in C++.
 * EPL plug-in authors should derive from EPLPlugin and use APAMA_DECLARE_EPL_PLUGIN(Class) macro to export it for use by the correlator.
 *
 * The following classes from com::softwareag::connectivity:: are also imported 
 * into this namespace for use with EPL plug-ins
 *
 *  - com::softwareag::connectivity::decimal_t;
 *  - com::softwareag::connectivity::data_t;
 *  - com::softwareag::connectivity::get;
 *  - com::softwareag::connectivity::list_t;
 *  - com::softwareag::connectivity::map_t;
 *  - com::softwareag::connectivity::buffer_t;
 *  - com::softwareag::connectivity::custom_t;
 *  - com::softwareag::connectivity::visitor;
 *  - com::softwareag::connectivity::const_visitor;
 *  - com::softwareag::connectivity::apply_visitor;
 *  - com::softwareag::connectivity::Logger;
 */
namespace _EPLPLUGIN_INTERNAL_CPP_NAMESPACE { 

// import these types into this namespace for convenience

using com::softwareag::connectivity::decimal_t;
using com::softwareag::connectivity::data_t;
using com::softwareag::connectivity::get;
using com::softwareag::connectivity::list_t;
using com::softwareag::connectivity::map_t;
using com::softwareag::connectivity::buffer_t;
using com::softwareag::connectivity::custom_t;
using com::softwareag::connectivity::visitor;
using com::softwareag::connectivity::const_visitor;
using com::softwareag::connectivity::apply_visitor;
using com::softwareag::connectivity::Logger;

}}} // com.apama.epl


#include <sag_internal/epl_plugin_wrap.hpp>

namespace com {
namespace apama {
namespace _EPLPLUGIN_INTERNAL_CPP_NAMESPACE { 

/**
 * Data structure used for declaring methods in the initialize function (EPLPlugin) to be exported to EPL.
 */
template<typename DERIVED>
class MethodData
{
public:
	/// @internal
	MethodData() {}
	/** Return the number of declared methods */
	size_t getNumMethods() { return methods.size(); }
	// @internal
	~MethodData();
	/**
	 * Declare a method using the default signature.
	 * Use this method if your arguments and return values are only:
	 *  - int64_t
	 *  - double
	 *  - const char *
	 *  - bool
	 *  - decimal_t
	 *  - custom_t<T>
	 * 
	 * @param name The name of the function to expose in EPL. This must be unique for a plug-in - there is no method overloading.
	 * @param blocking True if the function may block. False if it will never block. Default is true. May be overridden later with CorrelatorInterface::pluginMethodBlocking(). WARNING: blocking when declared as non-blocking can cause the correlator to deadlock
	 *
	 */
	template <typename FN_T, FN_T fn>
	void registerMethod(const char *name, bool blocking=true);
	/**
	 * Declare a method using a custom signature.
	 * Use this method if your arguments or return values contain:
	 *  - list_t
	 *  - map_t (either as an event type or a dictionary type)
	 *  - int64_t (for passing context objects)
	 * 
	 * @param name The name of the function to expose in EPL. This must be unique for a plug-in - there is no method overloading.
	 * @param signature The epl function type (in action variable signature form) of this function.
	 *
	 * An example signature would be:
	 * \code
	 * action<integer, sequence<float> > returns boolean
	 * \endcode
	 *
	 * @param blocking True if the function may block. False if it will never block. Default is true. May be overridden later with CorrelatorInterface::pluginMethodBlocking(). WARNING: blocking when declared as non-blocking can cause the correlator to deadlock
	 */
	template <typename FN_T, FN_T fn>
	void registerMethod(const char *name, const char *signature, bool blocking=true);
	/// @internal
	ApamaEPLMethodData *getMethods();
private:
	/// @internal
	std::vector<ApamaEPLMethodData> methods;
	MethodData(const MethodData &) = delete;
	MethodData &operator=(const MethodData &) = delete;
};

/**
 * This type is returned from CorrelatorInterface::registerEventHandler and can be used to add and remove channels
 * to the subscription for this event handler.
 */
class EventHandlerSubscription
{
	friend class CorrelatorInterface;
public:
	/** Create a dummy EventHandlerSubscription */
	EventHandlerSubscription(): internal(0), obj(0) {}
	/** Move an EventHandlerSubscription */
	EventHandlerSubscription(EventHandlerSubscription &&other)
		: internal(other.internal), obj(other.internal)
	{
		other.internal = 0;
		other.obj = 0;
	}
	/** Move an EventHandlerSubscription into this one */
	EventHandlerSubscription &operator=(EventHandlerSubscription &&other)
	{
		internal = other.internal;
		obj = other.obj;
		other.internal = 0;
		other.obj = 0;
		return *this;
	}
	/**
	 * Add a channel subscription to this event handler.
	 */
	void addChannel(const char *channel);
	/**
	 * Remove a channel subscription from this event handler.
	 * If the number of subscriptions is reduced to 0 this 
	 * handler will be deleted.
	 * Must not be called from within this class.
	 * @return true if the last channel was removed, otherwise false
	 */
	bool removeChannel(const char *channel);
	/**
	 * Remove all channels from this handler and delete it.
	 * Must not be called from within this class.
	 */
	void removeAllChannels();
private:
	/** @internal */
	EventHandlerSubscription(void *internal, void *obj): internal(internal), obj(obj) {}
	/** @internal */
	void *internal;
	/** @internal */
	void *obj;
	EventHandlerSubscription(const EventHandlerSubscription&) = delete;
	EventHandlerSubscription &operator=(const EventHandlerSubscription &) = delete;
};

/**
 * An interface for subscribing to events and receiving them in the plug-in.
 * Classes provide an instance of this to receive callbacks.
 * EventHandler registration is done by CorrelatorInterface::registerEventHandler.
 */
class EventHandler
{
public:
	/** Pointer to a type used to add and remove channels on a registered handler */
	typedef EventHandlerSubscription subscription_t;
	/** Pointer type for passing to registerEventHandler */
	typedef std::unique_ptr<EventHandler> ptr_t;
	/** To allow virtual deletion */
	virtual ~EventHandler() {}
	/** Implement this method to receive events.
	 * @param type The name of the event type
	 * @param event The event either in string form or in map form depending on what was specified at subscription type
	 * @param channel The channel the event is being delivered on.
	 */
	virtual void handleEvent(const char *type, data_t &&event, const char *channel) = 0;
};

/**
 * An interface for receiving notifications of threads being destroyed. 
 * This is typically used if you need to clean up any thread-local storage
 * used by the plug-in.
 * ThreadEnded callback registration is done by CorrelatorInterface::receiveThreadEndedCallbacks.
 */
class ThreadEndedHandler
{
public:
	/** Pointer type for passing to receiveThreadEndedCallbacks */
	typedef std::unique_ptr<ThreadEndedHandler> ptr_t;
	/** To allow virtual deletion */
	virtual ~ThreadEndedHandler() {}
	/** Called when a thread is to be destroyed, on the thread being destroyed */
	virtual void threadEnded() = 0;
};

/**
 * An interface to the correlator which lets plug-ins call methods on the host.
 * Can be obtained via calling com::apama::epl::EPLPlugin<DERIVED>::getCorrelator
 *
 * All of the methods on this interface are safe to be called from multiple threads.
 */
class CorrelatorInterface
{
public:
	/// @internal
	CorrelatorInterface(void *internal): internal(internal) {}
	/// @internal
	CorrelatorInterface(): internal(0) {}

	/** Mode for receiving events in.
	 * - STRING_MODE: as a string
	 * - MAP_MODE: as a map_t(string => data_t)
	 */
	enum event_mode_t
	{ 
		/** Deliver events in string form */
		STRING_MODE, 
		/** Deliver events in map_t(string => data_t) form */
		MAP_MODE
	};

	/** Register a handler to be called when a thread is about to be destroyed.
	 * Takes ownership of handler
	 */
	void receiveThreadEndedCallbacks(std::unique_ptr<ThreadEndedHandler> &&handler) const;

	/** 
	 * Register an event handler to receive events on the given channel in the given mode.
	 * @param handler The event handler to be registered (takes ownership).
	 * @param channel The channel to receive from.
	 * @param mode The mode to receive events in. Either STRING_MODE or MAP_MODE
	 * @param blocking True if the handler may block, False if it will not block (can be overridden with CorrelatorInterface::pluginMethodBlocking).
	 * @returns an object which can be used to add more channels to the subscription, or remove channels. When all channels are removed the handler will be deleted.
	 */
	EventHandler::subscription_t registerEventHandler(std::unique_ptr<EventHandler> &&handler, const char *channel, event_mode_t mode=MAP_MODE, bool blocking=true) const;
	/**
	 * Send an event in string form to a specific context.
	 * @param event The event in Apama string form
	 * @param ctx The ID of the context to deliver to 
	 * Context IDs can be obtained via CorrelatorInterface::getCurrentContextId (even if currently on another thread) or passed from EPL context objects, either directly or part of an event
	 */
	void sendEventTo(const std::string &event, uint64_t ctx) const
	{
		sendEventTo(event.c_str(), ctx);
	}
	/**
	 * Send an event in string form to a specific context.
	 * @param event The event in Apama string form
	 * @param ctx The ID of the context to deliver to 
	 * Context IDs can be obtained via CorrelatorInterface::getCurrentContextId (even if currently on another thread) or passed from EPL context objects, either directly or part of an event
	 */
	void sendEventTo(const char *event, uint64_t ctx) const;
	/**
	 * Send an event in map form to a specific context.
	 * @param type The name of the event type to parse this as
	 * @param event Either a map_t containing the event data as a string->object map or the event in Apama string form
	 * @param ctx The ID of the context to deliver to
	 * Context IDs can be obtained via CorrelatorInterface::getCurrentContextId (even if currently on another thread) or passed from EPL context objects, either directly or part of an event
	 */
	void sendEventTo(const char *type, data_t &&event, uint64_t ctx) const;
	/**
	 * Send an event in string form to a channel.
	 * @param event The event in Apama string form
	 * @param channel The channel to deliver to
	 */
	void sendEventTo(const std::string &event, const std::string &channel) const
	{
		sendEventTo(event.c_str(), channel.c_str());
	}
	/**
	 * Send an event in string form to a channel.
	 * @param event The event in Apama string form
	 * @param channel The channel to deliver to
	 */
	void sendEventTo(const char *event, const char *channel) const;
	/**
	 * Send an event in map form to a channel.
	 * @param type The name of the event type to parse this as
	 * @param event Either a map_t containing the event data as a string->object map or the event in Apama string form
	 * @param channel The channel to deliver to
	 */
	void sendEventTo(const char *type, data_t &&event, const char *channel) const;
	/**
	 * Indicate that a previously declared non-blocking method call or event handler may now actually be blocking.
	 * Method calls are marked as non-blocking via the blocking flag on the MethodData<DERIVED>::registerMethod function.
	 * Event handlers are marked as non-blocking via the blocking flag on the CorrelatorInterface::registerEventHandler method.
	 * This method is idempotent.
	 */
	void pluginMethodBlocking() const;
	/** 
	 * Return the ID of the current context, for passing to sendEventTo. 
	 * If being called from EPL, this returns the ID of the caller context.
	 * If called from own thread or from an event handler it will return 0, which is not legal to pass to sendEventTo.
	 */
	int64_t getCurrentContextId() const;
protected:
	/// @internal
	void *internal;
private:
	/// @internal
	static bool callEventHandler(void *handler, const char *type, sag_underlying_data_t event, const char *channel);
	/// @internal
	static void deleteEventHandler(void *handler);
	/// @internal
	static bool callThreadEndedHandler(void *handler);
	/// @internal
	static void deleteThreadEndedHandler(void *handler);

};

/**
 * Derive from this class to define a Correlator EPL plug-in.
 *
 * When you inherit from EPLPlugin you must template it over the derived class. Eg:
 *
 * \code
 * class MyPlugin: public EPLPlugin<MyPlugin> { ... };
 * \endcode
 *
 * Your class must implement zero-args constructor
 *
 * Your class must provide a static initialization method of this form:
 *
 * \code
 * static void initialize(base_plugin_t::method_data_t &md)
 * \endcode
 *
 * This method must export each of the methods which are to be visible to EPL using the MethodData<DERIVED>::registerMethod() function on method_data_t. Eg:
 *
 * \code
 * md.registerMethod<decltype(&MyPlugin::method), &MyPlugin::method>("eplMethodName");
 * \endcode
 *
 * This class provides the following members to its derived classes:
 *
 * - logger: a Logger object which can be used to log to the system log file.
 * - static getCorrelator(): a method which returns a CorrelatorInterface object which you can use to make callbacks into the correlator
 *
 * To export a class for use in the correlator you must use the following macro:
 *
 * APAMA_DECLARE_EPL_PLUGIN(Class)
 */
template<typename DERIVED>
class EPLPlugin
{
private:
	/// @internal
	typedef data_t (*call_method_fn_t)(DERIVED *self, const list_t &args);
public:
	/** Convenience typedef for the base type */
	typedef EPLPlugin base_plugin_t;
	/** Convenience typedef for the method data type */
	typedef MethodData<DERIVED> method_data_t;

	/** 
	 * Construct an EPL plug-in with a name (for logging)
	 * @param name The name of the plug-in
	 * */
	EPLPlugin(const std::string &name):
		logger("plugins."+name)
	{}
	/** Return a reference to the correlator host interface */
	static CorrelatorInterface &getCorrelator() { return correlator; }
	/// @internal
	static void setCorrelator(const CorrelatorInterface &corr) { correlator = corr; }
	/** 
	 * A Logger object which can be used to log to the correlator log file. 
	 * All log lines include the plug-in name specified in the EPLPlugin constructor.
	 * */
	Logger logger;

	/// @internal
	static data_t callMethod(void *self, void *fn, const list_t &args);
private:
	/// @internal
	DERIVED &derived() { return *static_cast<DERIVED*>(this); }
	/// The host interface
	static CorrelatorInterface correlator;
};

}
/**
 * Contains classes needed for implementing EPL plug-ins in C++.
 * EPL plug-in authors should derive from EPLPlugin and use APAMA_DECLARE_EPL_PLUGIN(Class) macro to export it for use by the correlator.
 *
 * The following classes from com::softwareag::connectivity:: are also imported 
 * into this namespace for use with EPL plug-ins
 *
 *  - com::softwareag::connectivity::decimal_t;
 *  - com::softwareag::connectivity::data_t;
 *  - com::softwareag::connectivity::get;
 *  - com::softwareag::connectivity::list_t;
 *  - com::softwareag::connectivity::map_t;
 *  - com::softwareag::connectivity::buffer_t;
 *  - com::softwareag::connectivity::custom_t;
 *  - com::softwareag::connectivity::visitor;
 *  - com::softwareag::connectivity::const_visitor;
 *  - com::softwareag::connectivity::apply_visitor;
 *  - com::softwareag::connectivity::Logger;
 */
namespace epl { using namespace _EPLPLUGIN_INTERNAL_CPP_NAMESPACE; }

}} // com.apama.epl

// internal implementation included from these files
#include <sag_internal/epl_plugin_functions.hpp>
#include <sag_internal/epl_plugin_methods.hpp>
#include <sag_internal/exception.hpp>
#include <sag_internal/epl_plugin_macros.hpp>

/** Creates the exported C free functions for an EPL plug-in to be loaded into the correlator.
 *
 * You must call this exactly once per shared library to export a plug-in class from that library.
 * The class provided must be derived from EPLPlugin<Class>.
 *
 * If your class is in a namespace then you must call this macro in the same namespace, passing in just the base name of the class.
 */
#define APAMA_DECLARE_EPL_PLUGIN(Class) _APAMA_DECLARE_EPL_PLUGIN(Class)

#endif // _APAMA_EPL_PLUGINS_HPP_
