/*
 * Title:        sag_internal/epl_plugins.hpp
 * Description:  Implementation of EPL plug-ins
 * $Copyright (c) 2015-2016, 2020, 2022 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 */


#ifndef _SAG_INTERNAL_EPL_PLUGIN_METHODS_HPP_
#define _SAG_INTERNAL_EPL_PLUGIN_METHODS_HPP_

namespace com {
namespace apama {
namespace _EPLPLUGIN_INTERNAL_CPP_NAMESPACE { 

/* ************************* Internal implementation of epl_plugin.hpp ************************************/

template<typename DERIVED>
CorrelatorInterface EPLPlugin<DERIVED>::correlator;

// ***************** MethodData implementation


template <typename DERIVED>
template <typename FN_T, FN_T fn>
inline void MethodData<DERIVED>::registerMethod(const char *name, const char *signature, bool blocking)
{
	ApamaEPLMethodData methodData;
	methodData.name = name;
	methodData.signature = strdup(signature);
	methodData.fn = Wrap<DERIVED, FN_T>::template void_star_of<fn>();
	methodData.blocking=blocking;
	methods.push_back(std::move(methodData));
}
template <typename DERIVED>
template <typename FN_T, FN_T fn>
inline void MethodData<DERIVED>::registerMethod(const char *name, bool blocking)
{
	ApamaEPLMethodData methodData;
	methodData.name = name;
	methodData.signature = Wrap<DERIVED, FN_T>::signature();
	methodData.fn = Wrap<DERIVED, FN_T>::template void_star_of<fn>();
	methodData.blocking=blocking;
	methods.push_back(std::move(methodData));
}
template <typename DERIVED>
ApamaEPLMethodData *MethodData<DERIVED>::getMethods()
{
	ApamaEPLMethodData *cmethods = new ApamaEPLMethodData[methods.size()+1];
	std::copy(methods.begin(), methods.end(), cmethods);
	cmethods[methods.size()].fn = 0;
	methods.clear();
	return cmethods;
}

template <typename DERIVED>
MethodData<DERIVED>::~MethodData()
{
	// in case of exceptions
	for (auto it = methods.begin(); it != methods.end(); ++it) {
		free(it->signature);
	}
}

// ************************** CorrelatorInterface implementation

inline void CorrelatorInterface::pluginMethodBlocking() const
{
	ap_plugin_method_blocking(internal);
}

inline void CorrelatorInterface::sendEventTo(const char *event, const char *channel) const
{
	sag_underlying_data_t evt;
	new (&evt) data_t(event);
	if (!ap_send_data(internal, "", evt, channel)) {
		throw std::runtime_error("Failed to send event to channel "+std::string(channel));
	}
}

inline void CorrelatorInterface::sendEventTo(const char *type, data_t &&event, const char *channel) const
{
	sag_underlying_data_t evt;
	new (&evt) data_t(std::move(event));
	if (!ap_send_data(internal, type, evt, channel)) {
		throw std::runtime_error("Failed to send event to channel "+std::string(channel));
	}
}

inline void CorrelatorInterface::sendEventTo(const char *event, uint64_t ctx) const
{
	sag_underlying_data_t evt;
	new (&evt) data_t(event);
	if (!ap_send_data_ctx(internal, "", evt, ctx)) {
		throw std::runtime_error("Failed to send event to context");
	}
}

inline void CorrelatorInterface::sendEventTo(const char *type, data_t &&event, uint64_t ctx) const
{
	sag_underlying_data_t evt;
	new (&evt) data_t(std::move(event));
	if (!ap_send_data_ctx(internal, type, evt, ctx)) {
		throw std::runtime_error("Failed to send event to context");
	}
}

inline void EventHandlerSubscription::addChannel(const char *channel)
{
	if (!internal) throw std::runtime_error("This is a dummy EventHandlerSubscription");
	if (!ap_add_channel_eventhandler(internal, obj, channel)) {
		throw std::runtime_error("Failed to add channel to event handler");
	}
}

inline bool EventHandlerSubscription::removeChannel(const char *channel)
{
	if (!internal) throw std::runtime_error("This is a dummy EventHandlerSubscription");
	if (ap_remove_channel_eventhandler(internal, obj, channel)) {
		obj = 0;
		return true;
	} else {
		return false;
	}
}

inline void EventHandlerSubscription::removeAllChannels()
{
	if (!internal) throw std::runtime_error("This is a dummy EventHandlerSubscription");
	ap_remove_channel_eventhandler(internal, obj, 0);
	obj = 0;
}

inline EventHandlerSubscription CorrelatorInterface::registerEventHandler(std::unique_ptr<EventHandler> &&_handler, const char *channel, event_mode_t mode, bool blocking) const
{
	EventHandler *handler = _handler.release();
	if (!ap_register_plugin_eventhandler(internal, handler, channel, mode, &callEventHandler, &deleteEventHandler, blocking)) {
		throw std::runtime_error("Failed to register EventHandler");
	}
	return EventHandlerSubscription(internal, handler);
}

inline bool CorrelatorInterface::callEventHandler(void *handler, const char *type, sag_underlying_data_t event, const char *channel)
{
	data_t evt{std::move(reinterpret_cast<data_t&>(event))};
	try {
		reinterpret_cast<EventHandler*>(handler)->handleEvent(type, std::move(evt), channel);
	} catch (const std::exception &e) {
		sag_log_at("EPLPlugin", SAG_LOG_ERROR, "Caught exception in EPL plugin event handler: %s", e.what());
		return false;
	} catch (...) {
		sag_log_at("EPLPlugin", SAG_LOG_ERROR, "Caught unknown exception in EPL plugin event handler");
		return false;
	}
	return true;
}

inline void CorrelatorInterface::deleteEventHandler(void *handler)
{
	delete reinterpret_cast<EventHandler*>(handler);
}

inline void CorrelatorInterface::receiveThreadEndedCallbacks(std::unique_ptr<ThreadEndedHandler> &&handler) const
{
	if (!ap_register_plugin_threadhandler(internal, handler.release(), &callThreadEndedHandler, &deleteThreadEndedHandler)) {
		throw std::runtime_error("Failed to register ThreadEndedHandler");
	}
}

inline bool CorrelatorInterface::callThreadEndedHandler(void *handler)
{
	try {
		reinterpret_cast<ThreadEndedHandler*>(handler)->threadEnded();
	} catch (const std::exception &e) {
		sag_log_at("EPLPlugin", SAG_LOG_ERROR, "Caught exception in EPL plugin thread-ended handler: %s", e.what());
		return false;
	} catch (...) {
		sag_log_at("EPLPlugin", SAG_LOG_ERROR, "Caught unknown exception in EPL plugin thread-ended handler: %s");
		return false;
	}
	return true;
}

inline void CorrelatorInterface::deleteThreadEndedHandler(void *handler)
{
	delete reinterpret_cast<ThreadEndedHandler*>(handler);
}

inline int64_t CorrelatorInterface::getCurrentContextId() const
{
	return ap_get_contextid(internal);
}


// *************************** EPLPlugin implementation

namespace detail {
/// casts a void* to a function pointer without warning.
/// WARNING: DRAGONS!
/// This is very non-portable, but portable enough and required when dlsyming a 
/// function.
template<typename FN>
inline FN unsafe_function_cast(void *const & vptr)
{
	return reinterpret_cast<FN const &>(vptr);
}
} // anonymous


template<typename DERIVED>
inline data_t EPLPlugin<DERIVED>::callMethod(void *self, void *fn, const list_t &args)
{
	return detail::unsafe_function_cast<call_method_fn_t>(fn)(reinterpret_cast<DERIVED*>(self), args);
}

}}} // com.apama.epl

#endif // _SAG_INTERNAL_EPL_PLUGIN_METHODS_HPP_
